package dont.panic.renewed.models;

import java.util.ArrayList;


import com.google.android.gms.maps.model.MarkerOptions;

import dont.panic.renewed.models.blockade.Blockade;
import dont.panic.renewed.models.equipment.Equipment;
import dont.panic.renewed.models.equipment.carryable.blockadetype.EquipmentForBlockade;
import dont.panic.renewed.models.equipment.noncarryable.FireHose;
import dont.panic.renewed.utils.EquipAndBlockNames;

/**
 * The object That represent the zone.
 *
 */
public class Zone {

	/**
	 * The name of the zone, which will be displayed at the map GUI
	 */
	private String name;
	
	/**
	 * The longitude of the Zone
	 */
	private double longitude;
	
	/**
	 * The latitude of the zone
	 */
	private double latitude;
	
	/**
	 * The list of all the blockades that are blocking the entrance (atm a zone can only be blocked by one blockade)
	 */
	private ArrayList<Blockade> blockadeList;
	
	/**
	 * The map marker of the zone
	 */
	private MarkerOptions marker;
	
	/**
	 * The panic level of the zone, this is used to tell which color the map marker should have
	 */
	private int panicLevel;
	
	/**
	 * The list of all the civilians that are within the zone
	 */
	private ArrayList<Civilian> civilianList;
	
	/**
	 * A boolean telling if its a normal zone or not. This is used to get the correct user interface up when a player has entered a zone.
	 * if its true its a normal zone with blockade and civilian, if its false then its the secure camp
	 */
	private boolean normalZone;
	
	/**
	 * A list of equipment which are within the zone 
	 */
	private ArrayList<Equipment> equipmentInZoneList;
	
	/**
	 * A list of strings recieved from the server which must be looped trough and real equipment object must be created.
	 */
	private ArrayList<String> equipmentFromServer;


	/**
	 * The constructor of the object, will initialise needed variables and set it coordination and name
	 * @param lat the latitude the zone will have
	 * @param lng the longitude the zone will have
	 * @param name the name of the zone which will be displayed at the MAP GUI
	 */
	public Zone(double lat, double lng, String name){
		this.longitude = lng;
		this.latitude = lat;
		this.name = name;
		this.blockadeList = new ArrayList<Blockade>();
		this.civilianList = new ArrayList<Civilian>();
		equipmentInZoneList = new ArrayList<Equipment>();
		panicLevel = 0;
		setUpTestBlockades();
		normalZone = true;
		setUpTestEquipmentInZone();
	}

	/**
	 * A test function used by us to check the various blockades and that the user interface of the blockade is correct
	 */
	private void setUpTestBlockades(){
		Blockade blockade = new Blockade(EquipAndBlockNames.FIRE,"The entrance is engulfed in fire");
		blockadeList.add(blockade);
		for(int i = 0; i < 7; i++){
			Civilian civ = new Civilian("test" +i);
			civilianList.add(civ);
		}
	}
	
	/**
	 * A test function used by us to check that the equipment within the zone is accessible
	 */
	private void setUpTestEquipmentInZone(){
		Equipment eq = new FireHose();
		equipmentInZoneList.add(eq);
	}
	
	/**
	 * A simple getter of the longitude
	 * @return the longitude of the zone
	 */
	public double getLongitude() {
		return longitude;
	}
	
	/**
	 * A setter which set the longitude of the zone
	 * @param longitude the longitude the zone should have
	 */
	public void setLongitude(double longitude) {
		this.longitude = longitude;
	}
	
	/**
	 * A simple getter which return the latitude of the zone.
	 * @return the latitude of the zone
	 */
	public double getLatitude() {
		return latitude;
	}
	
	/**
	 * Set the latitude of the zone
	 * @param latitude
	 */
	public void setLatitude(double latitude) {
		this.latitude = latitude;
	}
	
	/**
	 * A check to see if the player is nearby the zone
	 * @param lat the latitude of the zone
	 * @param lng the longitude of the zone
	 * @return true if the player is nearby, false if hes not nearby
	 */
	public boolean isWithinZone(double lat, double lng){
		double radius = 0.0002; //length given in latitude/longitude cord. equal about 10m
		
		if(Math.pow((lat - this.latitude), 2) + Math.pow((lng - this.longitude), 2) < Math.pow(radius, 2))
			return true; 	
		
		return false;

	}


	/**
	 * Go trough the blockades and check if they are unlocked or not.
	 * if all off them are unlocked, this means that this zone is unlocked as well
	 * @return true if the zone is locked, false if its unlocked.
	 */
	public boolean isLocked() {
		boolean locked = false;
		for (Blockade blockade : blockadeList) {
			if(blockade.isLocked()){
				locked = true;
				break;
			}
		}
		
		return locked;
		
	}


	/**
	 * Return the list of the blockade within the zone (atm its only 1 per zone=
	 * @return the list of blockades within the zone.
	 */
	public ArrayList<Blockade> getBlockadeList() {
		return blockadeList;
	}

	/**
	 * Get the name of the zone
	 * @return the name of the zone
	 */
	public String getName() {
		return name;
	}


	/**
	 * A simple setter which will set the name of the zone
	 * @param name The name of the zone.
	 */
	public void setName(String name) {
		this.name = name;
	}

	/**
	 * A simple getter which will return the marker of the zone, this is done when the marker is to change color
	 * @return the map marker representing this object.
	 */
	public MarkerOptions getMarker() {
		return marker;
	}
	/**
	 * A simple setter which will set the map marker used by this object.
	 * @param marker
	 */

	public void setMarker(MarkerOptions marker) {
		this.marker = marker;
	}
	
	/**
	 * increase the paniclevel of the zone
	 * @param amount
	 */
	public void increasePanic(int amount){
		panicLevel = panicLevel + amount;
	}

	/**
	 * Returns the list of civilians within the zone
	 * @return the list of civilians that are within the zone.
	 */
	public ArrayList<Civilian> getCivilianList() {
		return civilianList;
	}

	/**
	 * A setter which set if the zone is a normal zone or a secure camp.
	 * @param val
	 */
	public void setNormalZone(Boolean val) {
		this.normalZone = val;
	}
	
	/**
	 * Returns wheter this is a normal zone or the secure camp
	 * @return true if its a normal zone, false if its the secure camp
	 */
	public boolean isNormalZone(){
		return this.normalZone;
	}
	
	/**
	 * This function is called when the player or team mate has unlocked a blockade in this zone.
	 * @param name
	 */
	public void unlockBlockade(String name){
		for(Blockade block : blockadeList){
			if(block.getName().equalsIgnoreCase(name)){
				block.unlock();
			}
		}
	}
	
	/**
	 * This function is called when the player has performed an action on a civilian or when the team mate has done so.
	 * @param civName The name of the civilian which the action is to be performed on
	 * @param actionName the name of the action (calm, healing)
	 */
	public void performActionOnCivilian(String civName, String actionName){
		for(Civilian civ : civilianList){
			if(civ.getName().equalsIgnoreCase(civName)){
				civ.applyAction(actionName);
				break;
			}
				
		}
	}
	
	/**
	 * This function is called when a player has used an equipment directed toward the civilians in the zone.
	 * Note that this can be the player or a team mate.
	 * @param itemName the name of the item
	 */
	public void performActionOnAllCivilian(String itemName){
		for(Civilian civ : civilianList){
			civ.applyEquipment(itemName);
		}
	}

	
	/**
	 * A simple getter that will return the equipment which is within the zone
	 * @return the list of equipment which is within the zone
	 */
	public ArrayList<Equipment> getEquipmentInZoneList() {
		return equipmentInZoneList;
	}
	
	
	/**
	 * This function is called when a panic wave network message is recieved. It will increase the panic level of all the civilians with 10.
	 */
	public void panickWave(int amount){
		for(Civilian civ : civilianList){
			civ.addPanic(amount);
		}
		calculatePanicLevel();
	}
	
	/**
	 * The function that calculate the panic level of the zone by taking the avg panic of all the civilians
	 */
	public void calculatePanicLevel(){
		int max =1;
		for(Civilian civ : civilianList){
			if(civ.isAlive())
				max = max+ civ.getPanic();
		}
		if(civilianList.size() > 0){
			panicLevel = max/civilianList.size();
		}
		else {
			panicLevel = 1;
		}
	}

	/**
	 * A simple getter that returns the panic level of the zone
	 * @return the panic level of the zone
	 */
	public int getPanicLevel() {
		return panicLevel;
	}
	
	/**
	 * This function is called when a game tick network message is sent from the server
	 * It will reduce the health of the civilians which has 100 panic with 10 
	 */
	public void gameTick(){
		for(Civilian civ : civilianList){
			civ.reduceHealth(10);
		}
	}
	
	/**
	 * This function create the correct equipment object from the strings with equipment name sent from the server
	 * This is done to keep the network packages as small as possible.
	 */
	public void createEquipmentFromServer(){
		if(equipmentFromServer != null){
			for(String name : equipmentFromServer){
				EquipmentForBlockade  eq = (EquipmentForBlockade) EquipAndBlockNames.getEquipmentByName(name); 
				if(eq != null)
					equipmentInZoneList.add(eq);
			}
		}
		equipmentFromServer = null;
	}
	

}
